/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#section common
/*
 * Similar to:
 * include/asm-ia64/system.h
 *
 * System defines. Note that this is included both from .c and .S
 * files, so it does only defines, not any C code.  This is based
 * on information published in the Processor Abstraction Layer
 * and the System Abstraction Layer manual.
 *
 * Copyright (C) 1998-2003 Hewlett-Packard Co
 *	David Mosberger-Tang <davidm@hpl.hp.com>
 * Copyright (C) 1999 Asit Mallick <asit.k.mallick@intel.com>
 * Copyright (C) 1999 Don Dugger <don.dugger@intel.com>
 */

#section marcel_macros

/*
 * taken from Linux (include/asm-powerpc/system.h)
 * Copyright (C) 1999 Cort Dougan <cort@cs.nmt.edu>
 */

static __tbx_inline__ unsigned long
__xchg_u32(volatile void *p, unsigned long val)
{
        unsigned long prev;

        __asm__ __volatile__(
	"lwsync	\n"
"1:     lwarx   %0,0,%2 \n"
"       stwcx.  %3,0,%2 \n\
        bne-    1b"
	"\n\tisync\n"
        : "=&r" (prev), "+m" (*(volatile unsigned int *)p)
        : "r" (p), "r" (val)
        : "cc", "memory");

        return prev;
}



static __tbx_inline__ unsigned long
__xchg_u32_local(volatile void *p, unsigned long val)
{
        unsigned long prev;

        __asm__ __volatile__(
"1:     lwarx   %0,0,%2 \n"
"       stwcx.  %3,0,%2 \n\
        bne-    1b"
        : "=&r" (prev), "+m" (*(volatile unsigned int *)p)
        : "r" (p), "r" (val)
        : "cc", "memory");

        return prev;
}


static __tbx_inline__ unsigned long
__xchg_u64(volatile void *p, unsigned long val)
{
        unsigned long prev;

        __asm__ __volatile__(
	"lwsync	\n"
"1:     ldarx   %0,0,%2 \n"
"       stdcx.  %3,0,%2 \n\
        bne-    1b"
	"\n\tisync\n"
        : "=&r" (prev), "+m" (*(volatile unsigned long *)p)
        : "r" (p), "r" (val)
        : "cc", "memory");

        return prev;
}


static __tbx_inline__ unsigned long
__xchg_u64_local(volatile void *p, unsigned long val)
{
        unsigned long prev;

        __asm__ __volatile__(
"1:     ldarx   %0,0,%2 \n"
"       stdcx.  %3,0,%2 \n\
        bne-    1b"
        : "=&r" (prev), "+m" (*(volatile unsigned long *)p)
        : "r" (p), "r" (val)
        : "cc", "memory");

        return prev;
}


static __tbx_inline__ unsigned long
__xchg(volatile void *ptr, unsigned long x, unsigned int size)
{
        switch (size) {
	        case 4:
	                return __xchg_u32(ptr, x);
	        case 8:
	                return __xchg_u64(ptr, x);
		default:
			abort();
        }
        return x;
}

static __tbx_inline__ unsigned long
__xchg_local(volatile void *ptr, unsigned long x, unsigned int size)
{
        switch (size) {
        	case 4:
        	        return __xchg_u32_local(ptr, x);
        	case 8:
        	        return __xchg_u64_local(ptr, x);
		default:
			abort();
        }
        return x;
}

#define ma_xchg(ptr,x)                                                          \
  ({                                                                         \
     __typeof__(*(ptr)) _x_ = (x);                                           \
     (__typeof__(*(ptr))) __xchg((ptr), (unsigned long)_x_, sizeof(*(ptr))); \
  })


#define xchg_local(ptr,x)                                                    \
  ({                                                                         \
     __typeof__(*(ptr)) _x_ = (x);                                           \
     (__typeof__(*(ptr))) __xchg_local((ptr),                                \
                (unsigned long)_x_, sizeof(*(ptr)));                         \
  })

/*
 * Macros to force memory ordering.  In these descriptions, "previous"
 * and "subsequent" refer to program order; "visible" means that all
 * architecturally visible effects of a memory access have occurred
 * (at a minimum, this means the memory has been read or written).
 *
 *   ma_wmb():	Guarantees that all preceding stores to memory-
 *		like regions are visible before any subsequent
 *		stores and that all following stores will be
 *		visible only after all previous stores.
 *   ma_rmb():	Like wmb(), but for reads.
 *   ma_mb():	wmb()/rmb() combo, i.e., all previous memory
 *		accesses are visible before all subsequent
 *		accesses and vice versa.  This is also known as
 *		a "fence."
 *
 * Note: "ma_mb()" and its variants cannot be used as a fence to order
 * accesses to memory mapped I/O registers.  For that, mf.a needs to
 * be used.  However, we don't want to always use mf.a because (a)
 * it's (presumably) much slower than mf and (b) mf.a is supported for
 * sequential memory pages only.
 */

#define ma_mb()		__asm__ __volatile__ ("sync" : : : "memory")
#define ma_rmb()	__asm__ __volatile__ ("lwsync" : : : "memory")
#define ma_wmb()	__asm__ __volatile__ ("sync" : : : "memory")
#define ma_read_barrier_depends()	do { } while(0)

#ifdef MA__LWPS
# define ma_smp_mb()	ma_mb()
# define ma_smp_rmb()	ma_rmb()
# define ma_smp_wmb()	__asm__ __volatile__ ("eieio" : : : "memory")
# define ma_smp_read_barrier_depends()	ma_read_barrier_depends()
#else
# define ma_smp_mb()	ma_barrier()
# define ma_smp_rmb()	ma_barrier()
# define ma_smp_wmb()	ma_barrier()
# define ma_smp_read_barrier_depends()	do { } while(0)
#endif

/*
 * XXX check on these---I suspect what Linus really wants here is
 * acquire vs release semantics but we can't discuss this stuff with
 * Linus just yet.  Grrr...
 */
#define ma_set_mb(var, value)	do { (var) = (value); ma_mb(); } while (0)
#define ma_set_wmb(var, value)	do { (var) = (value); ma_wmb(); } while (0)

/* Macros for adjusting thread priority (hardware multi-threading) */
#define HMT_very_low()   asm volatile("or 31,31,31   # very low priority")
#define HMT_low()>------ asm volatile("or 1,1,1>     # low priority")
#define HMT_medium_low() asm volatile("or 6,6,6      # medium low priority")
#define HMT_medium()>--- asm volatile("or 2,2,2>     # medium priority")
#define HMT_medium_high() asm volatile("or 5,5,5      # medium high priority")
#define HMT_high()>----- asm volatile("or 3,3,3>     # high priority")

#define ma_cpu_relax() do { HMT_low(); HMT_medium(); barrier(); } while (0)
