/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2008 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <stdio.h>
#include "mm_mami.h"
#include "mm_mami_private.h"

#if defined(MM_MAMI_ENABLED)

int main(int argc, char **argv) {
  mami_manager_t *memory_manager;
  int i, err;
  int minsource, maxsource, mindest, maxdest;

  common_init(&argc, argv, NULL);
  mami_init(&memory_manager);

  minsource = 0;
  maxsource = memory_manager->nb_nodes-1;
  mindest = 0;
  maxdest = memory_manager->nb_nodes-1;

  for(i=1 ; i<argc ; i++) {
    if (!strcmp(argv[i], "-src")) {
      if (i+1 >= argc) abort();
      minsource = atoi(argv[i+1]);
      maxsource = atoi(argv[i+1]);
    }
    else if (!strcmp(argv[i], "-dest")) {
      if (i+1 >= argc) abort();
      mindest = atoi(argv[i+1]);
      maxdest = atoi(argv[i+1]);
    }
  }

  err = mami_sampling_of_memory_access(memory_manager, minsource, maxsource, mindest, maxdest);
  if (err < 0) perror("mami_sampling_of_memory_migration");

  mami_exit(&memory_manager);
  common_exit(NULL);
  return err;
}

#else
int main(int argc, char * argv[]) {
  fprintf(stderr, "This application needs MaMI to be enabled\n");
}
#endif
