

/* pseudo fonction pour trouver o crire un vnement dans les buffers */

/*
 * Structure du buffer de trace
 *
 * ADDR (a<b<c...)
 * ||
 * \/
 * 
 * a /--------------- 1re zone
 *     struct record
 *        last_event = c
 *        start = a
 *        end = e
 *
 * b   ev1
 *
 *     ev2
 *
 *     ...
 *
 * c   evn
 *
 * d   FREE SPACE
 * 
 * e /--------------- 2ime zone
 *     struct record
 *        last_event = g
 *        start = e
 *        end = i
 *        
 * f   ev1
 *
 *     ...
 *
 * g   evn
 *
 * h   FREE SPACE
 *
 * i /--------------- 3ime zone
 *   ...
 * 
 * */


// Variables globales :
void *start_buffer;
void *pos_buffer;
void *end_buffer;

struct record {
	void* last_event;
	void* start;
	void* end;
}

// Variables par thread (ie faire toto=th_buff revient  faire un
// get_specifique ou quivalent):
__thread struct th_buff {
	void* pos; /* pointe sur le FREE SPACE correspondant */
	struct record* infos; /* pointe sur le record de ce thread */
} th_buff;

#define SIZE_BUFF 200 /* au hazard */



int write_ev(size) {
  
	th_buff old_th, new_th;
  
  restart:
 	/* lecture atomique des 2 valeurs (8 octets) */
 	read8(th_buff, old_th);

	new_th = old_th;
	new_th.pos += size;

	if (new_th.pos > new_th.infos->end) {
     		/* les infos ne tiennent pas dans la place du buffer 
		 * On cherche un nouveau buffer/record */
		void* old_pos = pos_buffer;
		
		void* new_pos = old_pos + SIZE_BUFF;

		if (new_pos > end_buffer) {
			/* On stoppe tout comme actuellement */
			...

			return -E_NOSPACELEFT;
		}
		if (!lock_cmpxchg(pos_buffer, old_pos, new_pos)) {
			/* Des choses ont changs, peut-tre nous qui avons t
			 * interrompu et avons alors dj rexcut ce code
			 * (donc donn une nouvelle zone  ce thread */
			goto restart;
		}
		/* Ok, on est propritaire de la nouvelle rgion qui va de
		 * old_pos  new_pos */
		
		/* On prvoit de la place pour :
		 * 1) les donnes de controle (struct record)
		 * 2) notre vnement
		 * */
		new_th.pos=old_pos+sizeof(record_t)+size;
		
		*((struct record*)old_pos) = {
			.start=old_pos;
			.end=new_pos;
			.last_event=old_pos;
		};
		
		/* Sur ia32, on a cmpxchg8,
		 * mais sur ia64, on a seulement cmp4xchg8 (en fait cmp8xchg16
		 * vu qu'on est en 64 bits)
		 * J'cris donc l'algo avec le plus contraignant
		 *
		 * cmp4xchg8 : compare 4 octets, mais en cas d'galit on en crit 8
		 * */
		if (!cmp4xchg8(th_buff.pos, old_th.pos, new_th)) {
			/* Arghhh, on a t interrompu par une autre mesure qui
			 * a mis en place un nouveau buffer
			 * */
			if (lock_cmpxchg(pos_buffer, new_pos, old_pos)) {
				/* On redonne la zone qu'on vient d'allouer si
				 * c'est encore possible */
				goto restart;
			}
			/* Bon, ben on a une zone avec une mesure et pas grand
			 * chose d'autre  mettre dedans...
			 * Tant pis
			 *
			 * cas (*) (voir TODO)
			 * */
		}
	}
	/* criture des donnes  l'adresse :
	 * new_th.pos - size
	 * */
	...
  restart_update:
	/* Et mise  jour du pointeur sur la dernire criture de la zone */
	int old_last_ev=new_th.infos->last_event;
	if (new_th.pos > old_last_ev) {
		if (!cmpxchg(new_th.infos->last_event, old_last_ev, new_th.pos)) {
			goto restart_update;
		}
	}
	return 0;
}


TODO :
* Grer les FREE SPACE abandonnes
  En faire une liste chane
    - ajout dans cette fonction dans le cas (*)
    - ajout aprs la terminaison d'un thread qui laisse un grand FREE SPACE
    - consommation dans cette fonction avant de tenter d'allouer une nouvelle zone
      dans le buffer global

* Dumper tous les ev sur disque

* Faire en sorte que les trois variables globales et les deux variables par thread
  soit en fait des tableaux de variables index par la QBB courrante
  (on peut se planter de QBB parfois si on change de QBB entre le dbut de la
   fonction o on se renseigne et la fin de la fonction, mais ce n'est pas
   grave (juste une lgre perte de perf))

* BEAUCOUP plus dur :
  faire en sorte que le buffer principal puisse tre libr/sav/chang en
  cours d'excution

REMARQUE :
* en mettant le PID du thread dans la structure record, on n'a plus besoin de
  le mettre dans chaque vnement :-)
 
