/** @file
 ** @brief Base types and Puk management code.
 * @ingroup Puk
 */

  /** Initializes Puk */
  extern void padico_puk_init(int argc, char**argv);

  /** Start Puk bootstrap module */
  extern void padico_puk_start(void);

  /** Shutdowns Puk */
  extern void padico_puk_shutdown(void);

  /** shortcut for: padico_puk_init/start/shutdown, for simple padico-d bootstrap */
  extern void padico_puk_bootstrap(int argc, char**argv);

  /** Tests whether Puk is already initialized */
  extern int padico_puk_initialized(void);

  /* Data structures: lists, vectors, hashtables */

#define PUK_AUTO_TYPE(ENAME) \
  PUK_AUTO_TYPE_DECLARE(ENAME) \
  PUK_AUTO_TYPE_CREATE(ENAME)

#define PUK_AUTO_TYPE_DECLARE(ENAME) \
  typedef struct ENAME##_s * ENAME##_t;

#define PUK_AUTO_TYPE_CREATE(ENAME) \
  void ENAME##_ctor(ENAME##_t); \
  void ENAME##_dtor(ENAME##_t); \
  /** @internal */ \
  static inline ENAME##_t ENAME##_new(void) \
  { ENAME##_t obj=(ENAME##_t)padico_malloc(sizeof(struct ENAME##_s)); ENAME##_ctor(obj); return obj; } \
  /** @internal */ \
  static inline void ENAME##_delete(ENAME##_t obj) \
  { ENAME##_dtor(obj); padico_free(obj); }

  /* *** Puk args */

  /** Return a copy of argc */
  int padico_getargc(void);

  /** Return a pointer to the real argc.
   *  Use with care! */
  int* padico_getpargc(void);

  /** Return a copy of argv */
  char** padico_getargv(void);

  /** Return a pointer to the real argv.
   *  Use carefully */
  char*** padico_getpargv(void);

  /** Sets the value of an environment variable. */
  void padico_setenv(const char *var, const char *val);

  /** Returns the value of an environment variable. */
  const char*padico_getenv(const char*var);

  const char*padico_hostname(void);

  /** Returns the value of PADICO_ROOT */
  const char*padico_root(void);

  /** Abort the processus with a fatal error.
   * @param format Error message in "printf" format 
   * @warning Use with care! It aborts the *whole process*, not only the thread/module.
   */
  void padico_fatal(const char*format, ...) __attribute__ ((noreturn))  __attribute__ ((format (printf, 1, 2)));
  
  void padico_warning(const char*format, ...) __attribute__ ((format (printf, 1, 2)));

  void padico_print(const char*format, ...)  __attribute__ ((format (printf, 1, 2)));


  /* padico_malloc & co. Everything is managed in PukABI, so nothing to do now */
#define padico_malloc  malloc
#define padico_free    free
#define padico_realloc realloc
#define padico_calloc  calloc
#define padico_strdup  strdup
