/** @file
 * @brief Internal definitions for use inside Puk only.
 * @ingroup Puk
 */

/** @defgroup Puk Core module: Puk -- PadicoTM micro-kernel
 * @author Alexandre Denis
 */


#ifndef PUK_INTERNALS_H
#define PUK_INTERNALS_H

#include "Puk.h"

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

__PUK_SYM_INTERNAL void padico_puk_trace_init(void);
__PUK_SYM_INTERNAL void padico_puk_pkg_init(void);
__PUK_SYM_INTERNAL void padico_puk_lib_init(void);
__PUK_SYM_INTERNAL void padico_puk_monitor_init(void);
__PUK_SYM_INTERNAL void padico_puk_mod_init(void);
__PUK_SYM_INTERNAL void padico_puk_xml_init (void);
__PUK_SYM_INTERNAL void padico_puk_boot_init(void);
__PUK_SYM_INTERNAL void padico_puk_components_init(void);

__PUK_SYM_INTERNAL void padico_puk_trace_finalize(void);
__PUK_SYM_INTERNAL void padico_puk_xml_finalize(void);
__PUK_SYM_INTERNAL void padico_puk_components_finalize(void);
__PUK_SYM_INTERNAL void padico_puk_mod_finalize(void);


/* ** attrs */

struct padico_attr_s
{
  const char* label;
  const char* value;
};
typedef struct padico_attr_s*padico_attr_t;

/* ** units */
PUK_VECT_TYPE(padico_unit, struct padico_unit_s);

/* ** mod_refs */
PUK_VECT_TYPE(puk_mod, struct padico_mod_s*);

/* ** module */

struct padico_mod_s
{
  char*                     mod_name;    /**< name of the module */
  puk_adapter_t             driver;      /**< loader used to manage this module */
  struct padico_unit_vect_s units;       /**< units- content of the module */
  puk_hashtable_t           attrs;       /**< attributes of the module */
  padico_mod_t              inherits;    /**< ancestor for attributes inheriting */
  struct puk_mod_vect_s     used_by;     /**< modules using this one */
  struct puk_mod_vect_s     requires;    /**< modules used by this module */
  struct padico_modID_vect_s unresolved_deps; /**< unresolved requirements */
  int                       ref_count;   /**< number of references (other modules using this one or explicit 'load') */ 
};

#endif /* PUK_INTERNALS_H */
