/** @file
 * @brief Puk string-related stuff
 * @ingroup Puk
 */


/* *** Puk automatic strings ******************************* */

struct padico_string_s
{
  int   _size;
  int   _capacity;
  char* _text;
};
PUK_AUTO_TYPE_DECLARE(padico_string)

static inline void padico_string_ctor(padico_string_t s)
{
  static const int padico_string_default_capacity = 128;
  s->_capacity = padico_string_default_capacity;
  s->_text = (char*)padico_malloc(s->_capacity);
  s->_text[0] = 0;
  s->_size = 0;
}

static inline void padico_string_dtor(padico_string_t s)
{
  padico_free(s->_text);
}
PUK_AUTO_TYPE_CREATE(padico_string)

void padico_string_grow(padico_string_t s, int size_inc);
void padico_string_printf(padico_string_t s, const char*fmt, ...);
void padico_string_cat(padico_string_t s, const char*s2);
void padico_string_catf(padico_string_t s, const char*fmt, ...);

static inline char*padico_string_get(padico_string_t s)
{
  return s->_text;
}

/* *** Puk encoding **************************************** */

char*puk_escape_encode(const void*bytes, int*len, void*(*alloc)(unsigned long));
void*puk_escape_decode(const char*text,  int*len, void*(*alloc)(unsigned long));

char*puk_hex_encode(const void*bytes, int*len, void*(*alloc)(unsigned long));
void*puk_hex_decode(const char*text,  int*len, void*(*alloc)(unsigned long));



/* *** Puk error codes ************************************* */

/** @addtogroup PukMessage
 * @{
 */
/** Type used for return codes */
struct padico_rc_s
{
  int rc;              /**< the code itself */
  padico_string_t msg; /**< message (error or info) */
};
PUK_AUTO_TYPE_DECLARE(padico_rc)
/** @internal */
static inline void padico_rc_ctor(padico_rc_t rc)
{
  rc->rc = 0;
  rc->msg = NULL;
}
/** @internal */
static inline void padico_rc_dtor(padico_rc_t rc)
{
  if(rc && rc->msg)
    padico_string_delete(rc->msg);
}
PUK_AUTO_TYPE_CREATE(padico_rc)
static inline padico_rc_t padico_rc_ok(void) { return NULL; }
static inline int padico_rc_iserror(padico_rc_t rc) { return ((rc != NULL) && (rc->rc)); }
padico_rc_t padico_rc_create(int _rc, char*fmt, ...);
#define padico_rc_error(...) padico_rc_create(-1, __VA_ARGS__)
#define padico_rc_msg(...)   padico_rc_create(0, __VA_ARGS__)
padico_rc_t padico_rc_cat(padico_rc_t, padico_rc_t);
padico_rc_t padico_rc_add(padico_rc_t rc1 , padico_rc_t rc2);
void        padico_rc_show(padico_rc_t);
char*       padico_rc_gettext(padico_rc_t);
/** @} */

