#include "padico_config.h"

#ifdef PADICO_ENABLE_PUKABI_FSYS

/** File descriptor translation from ABI to real/virtual numbering
 * @ingroup Puk
 */

#include <unistd.h>
#include <poll.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Puk-ABI.h"


/* ********************************************************* */
/* ** fd_set */

int puk_fdset_isreal(const void*fds1, const void*fds2, const void*fds3, int fdmax)
{
  int isreal = -1;
  const fd_set*f1 = fds1;
  const fd_set*f2 = fds2;
  const fd_set*f3 = fds3;
  int i;
  for(i = 0; i < fdmax; i++)
    {
      if((f1 && FD_ISSET(i, f1)) ||
	 (f2 && FD_ISSET(i, f2)) ||
	 (f3 && FD_ISSET(i, f3)))
	{
	  if(isreal != -1 && puk_fds_isreal(i) != isreal)
	    {
	      while(1)
		;

	      puk_abi_fatal("PukABI: inconsistency detected while translating fdset at fd #%d (isreal=%d; fd_isreal=%d)\n",
			    i, isreal, puk_fds_isreal(i));
	    }
	  isreal = puk_fds_isreal(i);
	}
    }
  return isreal;
}

void puk_fdset_from_abi(void*fdset, int fdmax)
{
  int i;
  fd_set fdin;
  fd_set*fdout = fdset;
  if(fdset != NULL)
    {
      memcpy(&fdin, fdset, sizeof(fd_set));
      FD_ZERO(fdout);
      for(i=0; i<fdmax; i++)
	{
	  if(FD_ISSET(i, &fdin))
	    {
	      FD_SET(puk_fds_from_abi(i), fdout);
	    }
	}
    }
}

void puk_fdset_to_abi(void*fdset, int fdmax, int isreal)
{
  int i;
  fd_set fdin;
  fd_set*fdout = fdset;
  if(fdset != NULL)
    {
      memcpy(&fdin, fdset, sizeof(fd_set));
      FD_ZERO(fdout);
      for(i = 0; i < fdmax; i++)
	if(FD_ISSET(i, &fdin)) { FD_SET(puk_fds_lookup(i, isreal?1:2), fdout); }
    }
}


/* ********************************************************* */
/* ** pollfd */

void puk_pollfd_from_abi(void*ptr, int nfds)
{
  int i;
  struct pollfd* fds = ptr;
  for(i = 0; i < nfds; i++)
    {
      if(puk_fds_isreal(fds[i].fd))
	{
	  puk_fds_dump();
	  puk_abi_fatal("Puk-ABI: poll() cannot handle real fds (fd=%d).\n", fds[i].fd);
	}
      fds[i].fd = puk_fds_from_abi(fds[i].fd);
    }
}

void puk_pollfd_to_abi(void*ptr, int nfds)
{
  int i;
  struct pollfd* fds = ptr;
  for(i = 0; i < nfds; i++)
    {
      fds[i].fd = puk_fds_lookup(fds[i].fd, 2); /* always virtual */
      if(fds[i].fd < 0)
	{
	  puk_fds_dump();
	  puk_abi_fatal("Puk-ABI: poll() inconsistency in reverse mapping.\n");
	}
    }
}

#endif /* PADICO_ENABLE_PUKABI_FSYS */
