
#include "padico_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>

#ifdef HAVE_ELF_H
#include <elf.h>
#endif

#ifdef sun
#include <sys/elf.h>
#include <sys/link.h>
#include <sys/machelf.h>
#endif

struct needed_tab
{
  char *soname;
  int type;
};

char *readsoname(char *name, FILE *infile) {
  Elf32_Ehdr  *epnt;
  Elf32_Phdr *ppnt;
  int i;
  char *header;
  unsigned int dynamic_addr = 0;
  unsigned int dynamic_size = 0;
  int strtab_val = 0;
  int needed_val;
  int loadaddr = -1;
  Elf32_Dyn *dpnt;
  struct stat st;
  char *needed;
  char *soname = NULL;
  
  
  if (fstat(fileno(infile), &st))
    return NULL;
  header = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fileno(infile), 0);
  if (header == (caddr_t)-1)
    return NULL;
  
  epnt = (Elf32_Ehdr *)header;
  if ((int)(epnt+1) >
      (int)(header + st.st_size))
    goto skip;
  
  ppnt = (Elf32_Phdr *)&header[epnt->e_phoff];
  if ((int)ppnt < (int)header ||
      (int)(ppnt+epnt->e_phnum) > (int)(header + st.st_size))
    goto skip;
  
  for(i = 0; i < epnt->e_phnum; i++)
    {
      if (loadaddr == -1 && ppnt->p_type == PT_LOAD) 
	loadaddr = (ppnt->p_vaddr & 0xfffff000) - (ppnt->p_offset & 0xfffff000);
      if(ppnt->p_type == PT_DYNAMIC)
	{
	  dynamic_addr = ppnt->p_offset;
	  dynamic_size = ppnt->p_filesz;
	};
      ppnt++;
    };
  
  dpnt = (Elf32_Dyn *) &header[dynamic_addr];
  dynamic_size = dynamic_size / sizeof(Elf32_Dyn);
  if ((int)dpnt < (int)header ||
      (int)(dpnt+dynamic_size) > (int)(header + st.st_size))
    goto skip;
  
  for (i = 0; i < dynamic_size; i++)
    {
      if (dpnt->d_tag == DT_STRTAB)
	strtab_val = dpnt->d_un.d_val;
      dpnt++;
    };
  
  if (!strtab_val)
    goto skip;
  
  dpnt = (Elf32_Dyn *) &header[dynamic_addr];
  for (i = 0; i < dynamic_size; i++)
    {
      if (dpnt->d_tag == DT_SONAME || dpnt->d_tag == DT_NEEDED)
	{
	  needed_val = dpnt->d_un.d_val;
	  if (needed_val + strtab_val - loadaddr >= 0 ||
	      needed_val + strtab_val - loadaddr < st.st_size)
	    {
	      needed = (char *) (header + strtab_val + needed_val - loadaddr );
	      
	      if (dpnt->d_tag == DT_SONAME) {
		soname = strdup(needed);
	      }
	      
	    }
	}
      dpnt++;
    };
    
 
 skip:
  munmap(header, st.st_size);
  
  return soname;
}

int main(int argc, char *argv[]) {
  char null_str[]="(null)";
  FILE *f;
  char *name=argv[1];
  char *n;
  if ((f=fopen(name, "r")) == NULL) {
    fprintf(stderr,"Not found file %s\n",name);
    exit(1);
  }
  n=readsoname(name,f);
  if (n==NULL) {
    n=null_str;
  }
  printf("%s\n",n);
  return 0;
}
