#! /bin/bash

# ### site configuration

SVNURL="svn+ssh://padico.gforge.inria.fr/svn/padico/PadicoTM/trunk"
DISTRIB_DIR=${PADICO_ROOT}/var/tmp/release
VERSION=`cat VERSION`
RELEASE_NAME="PadicoTM-${VERSION}"
MAKE=make

# default add-ons included in "full" package
DISTRIB_ADDONS="MiddlewareServices/omniORB-4.0/package/omniORB-4.0.6.tar.gz MiddlewareServices/mpich-1.2.x/package/mpich-1.2.4-padico-svn20051027.tar.gz"

echo
echo "** Generating distrib for ${RELEASE_NAME}"
echo "   in ${DISTRIB_DIR}"
echo
mkdir -p ${DISTRIB_DIR}

echo "- Exporting Padico from SVN to ${DISTRIB_DIR}"
svn export ${SVNURL} ${DISTRIB_DIR}/${RELEASE_NAME}

echo "- Generating copyright..."
/bin/sh admin/gen_copyright.sh
echo "- Adding copyright headers..."
find ${DISTRIB_DIR}/${RELEASE_NAME} -print | xargs -n 1 /bin/sh admin/add_copyright.sh

echo "- Generating configuration files..."
( cd ${DISTRIB_DIR}/${RELEASE_NAME}; ${MAKE} config ) 2> /dev/null

echo "- Generating release note..."
mv ${DISTRIB_DIR}/${RELEASE_NAME}/RELEASE_NOTE ${DISTRIB_DIR}/${RELEASE_NAME}/RELEASE_NOTE.old
cat > ${DISTRIB_DIR}/${RELEASE_NAME}/RELEASE_NOTE <<EOF
This is PadicoTM version ${VERSION}
  (packaged by `whoami`, `date`)

Add-on packages included in this release:
${DISTRIB_ADDONS}

`cat ${DISTRIB_DIR}/${RELEASE_NAME}/RELEASE_NOTE.old`

EOF
rm ${DISTRIB_DIR}/${RELEASE_NAME}/RELEASE_NOTE.old

echo "- Generating user's manual..."
( cd ${DISTRIB_DIR}/${RELEASE_NAME}; ${MAKE} -C doc manual )

if [ x${PADICO_ROOT} != x ]; then \
  echo "Copying Java archives..." ; \
  cp ${PADICO_ROOT}/lib/classes/PadicoRemote.jar ${DISTRIB_DIR}/${RELEASE_NAME}/PadicoControl/ ; \
fi

echo
echo "- Export REALSE_NOTE"
cp ${DISTRIB_DIR}/${RELEASE_NAME}/RELEASE_NOTE ${DISTRIB_DIR}/

echo
echo "- Export doc"
cp ${DISTRIB_DIR}/${RELEASE_NAME}/doc/Padico-manual.ps ${DISTRIB_DIR}/

echo
echo "- Packaging ${RELEASE_NAME}-base.tar.gz"
( cd ${DISTRIB_DIR} ; tar czf ${RELEASE_NAME}-base.tar.gz ${RELEASE_NAME} )

echo "- Copying add-ons"
mkdir -p ${DISTRIB_DIR}/${RELEASE_NAME}/packages
for p in ${DISTRIB_ADDONS}; do
  echo "  - package: $p..."
  cp -a ${p} ${DISTRIB_DIR}/${RELEASE_NAME}/packages/
done
( cd ${DISTRIB_DIR}/${RELEASE_NAME}/MiddlewareServices ; ./package-admin import )

echo
echo "- Packaging ${RELEASE_NAME}.tar.gz"
( cd ${DISTRIB_DIR} ; tar czf ${RELEASE_NAME}.tar.gz ${RELEASE_NAME} )

echo
echo "- Purging ${RELEASE_NAME}..."
( cd ${DISTRIB_DIR} ; rm -rf ${RELEASE_NAME} )

echo
echo "** PadicoTM base distribution ready in: ${DISTRIB_DIR}"
echo
ls -l ${DISTRIB_DIR}
echo

